<?php
/**
 * FuzeWorks Framework Administration Plugin.
 *
 * The FuzeWorks PHP FrameWork
 *
 * Copyright (C) 2013-2020 i15
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * @author    i15
 * @copyright Copyright (c) 2013 - 2020, i15. (https://i15.nl)
 * @license   https://opensource.org/licenses/MIT MIT License
 *
 * @since Version 1.3.0
 *
 * @version Version 1.3.0
 */

namespace Application\Controller;

use FuzeWorks\Core;
use FuzeWorks\Exception\NotFoundException;
use FuzeWorks\Priority;
use FuzeWorks\WebController;

class SettingsController extends WebController
{

    public function findSettingsFiles(): array
    {
        // First collect all the known paths
        $paths = [];
        for ($i = Priority::getHighestPriority(); $i <= Priority::getLowestPriority(); $i++)
            $paths[$i] = $this->config->getComponentPaths($i);

        // Then go over every individual paths
        $configFiles = [];
        foreach ($paths as $priority => $foldersArray) {
            foreach ($foldersArray as $folder) {
                // Get the contents from the folder
                $contents = array_diff(scandir($folder), array('..', '.'));

                // Then go over the folders, and see which ones are an admin view
                foreach ($contents as $file) {
                    // If the file matches the expected filename, add it to the list
                    if (str_starts_with($file, 'config.')) {
                        if (!isset($configFiles[$file]))
                            $configFiles[$file] = [];

                        $configFiles[$file][] = ['folder' => $folder, 'priority' => substr(Priority::getPriority($priority), 10)];
                    }
                }
            }
        }

        // At the end, sort the files
        ksort($configFiles);

        // And return them
        return $configFiles;
    }

    /**
     * Fetch a file
     *
     * @param string $fileName
     * @return array
     * @throws NotFoundException
     */
    public function fetchFile(string $fileName, int $selector): array
    {
        // Fetch files
        $files = $this->findSettingsFiles();

        // Check if the file exists
        if (!isset($files[$fileName]))
            throw new NotFoundException("The requested settings file could not be found.");

        // Select correct data
        $meta = $files[$fileName][$selector];

        // Find the file
        $file = $meta['folder'] . DS . $fileName;
        if (!file_exists($file))
            throw new NotFoundException("The requested settings file could not be found.");

        // Load the raw data
        $raw = file_get_contents($file);

        // And try to load the editable data
        $data = include($file);

        // Prepare output variable
        return ['name' => $fileName, 'directory' => $meta['folder'], 'raw' => $raw, 'data' => $data];
    }

    protected function verifyFileData(array $fileData): bool
    {
        foreach ($fileData as $key => $val) {
            switch (gettype($val)) {
                case 'array':
                case 'object':
                case 'resource':
                case 'unknown type':
                    return false;
                    break;
            }
        }

        return true;
    }

}